import pytest

from sypht.util import DEFAULT_REC_LIMIT, fetch_all_pages


def test_fetch_all_pages_can_fetch_one_page():
    # arrange
    page_size = 5

    def fetch_something(offset, pages=1):
        pages0 = pages - 1
        if offset > pages0:
            return []
        start = offset * page_size
        page = range(start, start + page_size)
        return list(page)

    # act
    page_iter = fetch_all_pages(name="test1", fetch_page=fetch_something)
    results = []
    for page in page_iter(pages=1):
        results += page

    # assert
    assert results == [0, 1, 2, 3, 4]


def test_fetch_all_pages_can_fetch_one_page_with_get_page():
    # arrange
    page_size = 5

    def fetch_something(offset, pages=1):
        pages0 = pages - 1
        if offset > pages0:
            return {"results": []}
        start = offset * page_size
        page = range(start, start + page_size)
        return {"results": list(page)}

    # act
    page_iter = fetch_all_pages(
        name="test1", fetch_page=fetch_something, get_page=lambda resp: resp["results"]
    )
    results = []
    for resp in page_iter(pages=1):
        results += resp["results"]

    # assert
    assert results == [0, 1, 2, 3, 4]


def test_fetch_all_pages_can_fetch_several_pages():
    # arrange
    page_size = 5

    def fetch_something(offset, pages=1):
        pages0 = pages - 1
        if offset > pages0:
            return []
        start = offset * page_size
        page = range(start, start + page_size)
        return list(page)

    # act
    page_iter = fetch_all_pages(name="test1", fetch_page=fetch_something)
    results = []
    for page in page_iter(pages=2):
        results += page

    # assert
    assert results == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_fetch_all_pages_never_ending():
    """Fail if fetch more than n pages."""

    # arrange
    def never_ending(*args, **kwargs):
        return [0, 1, 2]

    # act
    page_iter = fetch_all_pages(name="test1", fetch_page=never_ending)
    results = []
    with pytest.raises(Exception) as exc_info:
        for page in page_iter():
            results += page

    # assert
    assert f"more than the limit: {DEFAULT_REC_LIMIT}" in str(exc_info)


def test_fetch_with_rec_limit():
    # arrange
    page_size = 5

    def fetch_something(offset, pages=1):
        pages0 = pages - 1
        if offset > pages0:
            return []
        start = offset * page_size
        page = range(start, start + page_size)
        return list(page)

    # act
    page_iter = fetch_all_pages(name="test1", fetch_page=fetch_something, rec_limit=2)
    results = []
    with pytest.raises(Exception) as exc_info:
        for page in page_iter():
            results += page

    # assert
    assert f"fetched 5 records which is more than the limit: 2" in str(exc_info)


def test_fetch_all_pages_handle_error():
    # arrange
    def failing(*args, **kwargs):
        raise Exception("fetch error")

    # act
    page_iter = fetch_all_pages(name="test1", fetch_page=failing)
    results = []
    with pytest.raises(Exception) as exc_info:
        for page in page_iter():
            results += page

    # assert
    assert "fetch error" in str(exc_info.value.__cause__)
    assert "Failed fetching for test1" in str(exc_info)



import unittest

from sypht.client import SyphtClient


class CreateFileTests(unittest.TestCase):
    def setUp(self):
        self.sypht_client = SyphtClient()

    def test_create_file(self):
        with open("tests/sample_invoice.pdf", "rb") as f:

            response = self.sypht_client.create_file(file=("sample_invoice.pdf", f))
            self.assertTrue(response["status"], "RECIEVED")

    def test_create_file_with_data(self):
        with open("tests/sample_invoice.pdf", "rb") as f:

            response = self.sypht_client.create_file(
                file=("sample_invoice.pdf", f), data={"splitState": "created_by_split"}
            )
            self.assertTrue(response["status"], "RECIEVED")


if __name__ == "__main__":
    unittest.main()



import json
import unittest
import warnings
from datetime import datetime, timedelta
from unittest.mock import Mock, patch
from uuid import UUID, uuid4

import httpretty
import pytest

from sypht.client import SyphtClient


def validate_uuid4(uuid_string):
    try:
        val = UUID(uuid_string, version=4)
    except ValueError:
        return False

    return val.hex == uuid_string.replace("-", "")


class DataExtraction(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", category=ResourceWarning)
        self.sypht_client = SyphtClient()

    def test_with_wrong_fieldset(self):
        with self.assertRaises(Exception) as context:
            with open("tests/sample_invoice.pdf", "rb") as f:
                response = self.sypht_client.upload(
                    f,
                    [
                        "sypht.incorrect",
                    ],
                )
                self.assertIn(
                    "does not have permission to use fieldSet sypht.incorrect",
                    response["error"],
                )

        self.assertTrue("Request failed with status code" in str(context.exception))

    def test_data_extraction_1(self):
        with open("tests/sample_invoice.pdf", "rb") as f:
            fid = self.sypht_client.upload(f, ["invoices:2"])
            self.assertTrue(validate_uuid4(fid))

        results = self.sypht_client.fetch_results(fid)

        self.assertTrue(isinstance(results, dict))
        self.assertIn("invoice.dueDate", results)
        self.assertIn("invoice.total", results)
        self.assertIn("invoice.amountPaid", results)
        self.assertIn("invoice.amountDue", results)

    def test_data_extraction_2(self):
        with open("tests/sample_invoice.pdf", "rb") as f:
            fid = self.sypht_client.upload(f, products=["sypht.invoice", "sypht.bank"])
            self.assertTrue(validate_uuid4(fid))

        results = self.sypht_client.fetch_results(fid)

        self.assertTrue(isinstance(results, dict))
        self.assertIn("invoice.dueDate", results)
        self.assertIn("invoice.total", results)
        self.assertIn("invoice.amountPaid", results)
        self.assertIn("invoice.amountDue", results)
        self.assertIn("bank.accountNo", results)
        self.assertIn("bank.bsb", results)

    def test_parent_doc_id(self):
        parent_doc_id = uuid4()
        with open("tests/sample_invoice.pdf", "rb") as f:
            fid = self.sypht_client.upload(f, ["invoices"], parent_doc_id=parent_doc_id)
            self.assertTrue(validate_uuid4(fid))


class ReauthenticateTest(unittest.TestCase):
    def setUp(self):
        self.sypht_client = SyphtClient()
        self.init_access_token = str(self.sypht_client._access_token)
        self.assertFalse(self.sypht_client._is_token_expired())

    def test_no_reauthentication(self):
        # Test non-expired token doesn't require reauthentication
        self.sypht_client.get_company()
        self.assertEqual(self.init_access_token, self.sypht_client._access_token)

    def test_reauthentication(self):
        # Set auth expiry to 1 second ago to avoid mocking datetime
        self.sypht_client._auth_expiry = datetime.utcnow() - timedelta(seconds=1)
        self.assertTrue(self.sypht_client._is_token_expired())

        # Get request will auto-reauthenticate.
        self.sypht_client.get_company()
        self.assertFalse(self.sypht_client._is_token_expired())


class RetryTest(unittest.TestCase):
    """Test the global retry logic works as we expect it to."""

    @patch.object(SyphtClient, "_authenticate_v2", return_value=("access_token", 100))
    @patch.object(SyphtClient, "_authenticate_v1", return_value=("access_token2", 100))
    @httpretty.activate(verbose=True, allow_net_connect=False)
    def test_it_should_retry_n_times(self, auth_v1: Mock, auth_v2: Mock):
        # arrange
        self.count = 0

        def get_annotations(request, uri, response_headers):
            self.count += 1
            # 1 req + 3 retries = 4
            if self.count == 4:
                return [200, response_headers, json.dumps({"annotations": []})]
            return [502, response_headers, json.dumps({})]

        httpretty.register_uri(
            httpretty.GET,
            "https://api.sypht.com/app/annotations?offset=0&fromDate=2021-01-01&toDate=2021-01-01",
            body=get_annotations,
        )

        sypht_client = SyphtClient(base_endpoint="https://api.sypht.com")

        # act / assert
        response = sypht_client.get_annotations(
            from_date=datetime(
                year=2021, month=1, day=1, hour=0, minute=0, second=0
            ).strftime("%Y-%m-%d"),
            to_date=datetime(
                year=2021, month=1, day=1, hour=0, minute=0, second=0
            ).strftime("%Y-%m-%d"),
        )

        assert response == {"annotations": []}

    @patch.object(SyphtClient, "_authenticate_v2", return_value=("access_token", 100))
    @patch.object(SyphtClient, "_authenticate_v1", return_value=("access_token2", 100))
    @httpretty.activate(verbose=True, allow_net_connect=False)
    def test_retry_should_eventually_fail_for_50x(self, auth_v1: Mock, auth_v2: Mock):
        # arrange
        self.count = 0

        def get_annotations(request, uri, response_headers):
            self.count += 1
            return [502, response_headers, json.dumps({})]

        httpretty.register_uri(
            httpretty.GET,
            "https://api.sypht.com/app/annotations?offset=0&fromDate=2021-01-01&toDate=2021-01-01",
            body=get_annotations,
        )

        sypht_client = SyphtClient(base_endpoint="https://api.sypht.com")

        # act / assert
        with self.assertRaisesRegex(Exception, ".") as e:
            sypht_client.get_annotations(
                from_date=datetime(
                    year=2021, month=1, day=1, hour=0, minute=0, second=0
                ).strftime("%Y-%m-%d"),
                to_date=datetime(
                    year=2021, month=1, day=1, hour=0, minute=0, second=0
                ).strftime("%Y-%m-%d"),
            )

        assert self.count == 4, "should be 1 req + 3 retries"


if __name__ == "__main__":
    unittest.main()



import logging
from typing import Any, Callable, Iterator, List

DEFAULT_REC_LIMIT = 100_000


def fetch_all_pages(
    name: str,
    fetch_page: Callable[..., Any],
    get_page: Callable[..., List[Any]] = lambda x: x,
    rec_limit=DEFAULT_REC_LIMIT,
) -> Callable[..., Iterator[Any]]:
    """Returns an iterator that calls fetch_page with an offset that we increment by the number of pages fetched.  Stop if page returns empty list.

    :param fetch_page: a function that makes an api call to fetch a page of results (using zero-based offset)
    :param get_page: a function that extracts the page from the response which should be a list
    """

    # Enforce a default so that the loop will stop.
    if rec_limit is None:
        rec_limit = DEFAULT_REC_LIMIT

    def fetch_all_pages(*args, **kwargs) -> Iterator[Any]:
        page_count = 0
        recs = 0
        while True:
            page_count += 1
            if recs > rec_limit:
                # Don't want to DOS ourselves...
                raise Exception(
                    f"fetch_all_pages({name}): fetched {recs} records which is more than the limit: {rec_limit} .  Consider adding or adjusting a filter to reduce the total number of items fetched."
                )
            try:
                response = fetch_page(
                    *args,
                    **kwargs,
                    offset=page_count - 1,
                )
            except Exception as err:
                raise Exception(
                    f"Failed fetching for {name} for offset={page_count - 1} (page={page_count}) (records fetched so far:{recs}). Cause: {err}"
                ) from err
            try:
                page = get_page(response)
            except Exception as err:
                raise Exception(
                    f"get_page failed to extract page from response for {name} for offset={page_count - 1} (page={page_count}) (records fetched so far:{recs}). Cause: {err}"
                ) from err
            if len(page) == 0:
                break
            recs += len(page)
            logging.info(
                f"fetch_all_pages({name}): fetched page {page_count} (records={recs})"
            )
            yield response

    return fetch_all_pages
